from time import sleep


class Heartbeat:
    def __init__(self, interval: int):
        self.interval = interval
        self.__heartbeat_counter = 0

    def __iter__(self):
        while True:
            self.__heartbeat_counter += 1
            yield self.__heartbeat_counter
            sleep(self.interval)



import math
from datetime import datetime
from hashlib import sha256
from json import dumps
from enum import Enum


def sigmoid(x: float) -> float:
    return 1 / (1 + math.exp(-x))


class JobStatus(Enum):
    PENDING = 1
    ASSIGNED = 2
    DONE = 3


class Job:
    def __init__(
            self,
            name: str,
            payload: dict,
            created_at: datetime = None,
            status: JobStatus = JobStatus.PENDING,
            assigned_to: int = 0,
            result: dict = None
    ):
        if created_at is None:
            created_at = datetime.utcnow()
        self.name = name
        self.payload = payload
        self.created_at = created_at
        self.status = status
        self.assigned_to = assigned_to
        self.result = result

    def assign(self, node_id: int):
        self.assigned_to = node_id
        self.status = JobStatus.ASSIGNED

    def set_result(self, result: dict):
        self.result = result
        self.status = JobStatus.DONE

    @property
    def id(self) -> str:
        job_as_string = f'{self.name},{dumps(self.payload, sort_keys=True)}'.encode()
        return sha256(job_as_string).hexdigest()

    @property
    def vector(self) -> float:
        date_sigmoid = 1 - sigmoid(round(self.created_at.timestamp()))  # older the better
        date_wight = 5
        id_sigmoid = sigmoid(self.assigned_to)  # tie break
        id_wight = 5
        return (date_sigmoid * date_wight) + (id_sigmoid * id_wight)

    def __gt__(self, other):
        return self.vector > other.vector

    def __eq__(self, other):
        return self.vector == other.vector  # rarely happened

    def __str__(self):
        return f'''\
    id = {self.id},
    name = {self.name},
    payload = {self.payload},
    assigned = {self.assigned_to},
    status = {self.status},
    result = {self.result},
'''

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'name': self.name,
            'payload': self.payload,
            'created_at': self.created_at.isoformat(),
            'status': self.status.value,
            'assigned_to': self.assigned_to,
            'result': self.result,
        }

    @classmethod
    def from_dict(cls, job_dict: dict):
        job = cls(
            name=job_dict['name'],
            payload=job_dict['payload'],
            created_at=datetime.fromisoformat(job_dict['created_at']),
            status=JobStatus(job_dict['status']),
            assigned_to=job_dict['assigned_to'],
            result=job_dict['result'],
        )
        return job

    def __hash__(self):
        return hash(f'{self.id}:{self.created_at}:{self.status}:{self.assigned_to}:{self.result}')



from threading import Thread
from random import sample
from socket import socket, AF_INET, SOCK_STREAM
from time import sleep
from typing import List
from socketserver import ThreadingTCPServer, BaseRequestHandler
from json import loads, dumps

from ..state import State
from ..base_communication import BaseCommunication
from ..job import Job


__all__ = ['SocketCommunication']


def state_from_dict(state_dict: dict) -> State:
    jobs = {
        job_dict['id']: Job.from_dict(job_dict)
        for job_dict in state_dict['jobs']
    }
    return State(jobs=jobs)


def state_to_dict(state: State) -> dict:
    jobs_list = [job.to_dict() for job in state._jobs.values()]
    return {'jobs': jobs_list}


class EmptyPeersListError(Exception):
    pass


class RequestHandler(BaseRequestHandler):
    def handle(self):
        request_data = self.request.recv(100*1000)
        if not request_data:
            return
        request_dict = loads(request_data)

        if request_dict['port'] == 34688:  # Testing on localhost
            self.client_address = ('127.0.0.2', 34688)

        self.server.communication.peers.add((self.client_address[0], request_dict['port']))
        if request_dict['route'] == 'ips':
            json_peers = dumps(list(self.server.communication.peers))
            self.request.send(json_peers.encode())
        elif request_dict['route'] == 'pull_state':
            json_state = dumps(state_to_dict(self.server.communication.current_state))
            self.request.send(json_state.encode())


class SocketServer(ThreadingTCPServer):
    def __init__(self, communication, *args, **kwargs):
        self.communication = communication
        super().__init__(*args, **kwargs)


class SocketCommunication(BaseCommunication, Thread):

    SOCKET_TIMEOUT = 15

    def __init__(self, host='0.0.0.0', port=34687, pull_interval=10, bootstrap_nodes: list = None):
        if bootstrap_nodes is None:
            bootstrap_nodes = []
        self._run = True
        self._pull_interval = pull_interval

        self.server_address = (host, port)
        self.peers = set(bootstrap_nodes)  # bootstrap nodes
        self.pulled_states = []
        self.socket_server = SocketServer(self, self.server_address, RequestHandler)
        self.current_state = State()

        super().__init__(name='SocketCommunication', daemon=True)

        self.socket_server_thread = Thread(target=self.socket_server.serve_forever, daemon=True)

        self.socket_server_thread.start()
        self.start()
        print(f'started socket server on {self.server_address}...')

    def pull(self) -> List[State]:
        list_copy = self.pulled_states.copy()
        self.pulled_states.clear()
        return list_copy

    def broadcast(self, state: State):
        self.current_state = state

    def _random_peer(self) -> tuple:
        for try_count in range(10):
            random_peer = sample(list(self.peers), 1)
            if random_peer[0] == self.server_address:
                continue
            return random_peer[0]
        raise EmptyPeersListError()

    def _update_ips(self):
        peer_address = self._random_peer()
        sock = socket()
        sock.settimeout(self.SOCKET_TIMEOUT)
        sock.connect(peer_address)
        sock.send(dumps({'route': 'ips', 'port': self.server_address[1]}).encode())
        ips_json = sock.recv(1000*1000)
        ips = map(tuple, loads(ips_json))
        self.peers |= set(ips)

    def _pull_state(self):
        peer_address = self._random_peer()
        sock = socket(AF_INET, SOCK_STREAM)
        sock.settimeout(self.SOCKET_TIMEOUT)
        sock.connect(peer_address)
        sock.send(dumps({'route': 'pull_state', 'port': self.server_address[1]}).encode())
        state_json = sock.recv(1000 * 1000)
        state = state_from_dict(loads(state_json))
        self.pulled_states.append(state)

    def run(self) -> None:
        errors = 0
        while self._run:
            sleep(self._pull_interval)
            if not self._run:
                break
            try:
                self._update_ips()
                self._pull_state()
            except ConnectionError as CE:
                print('connection error', CE)
                errors += 1
                continue
            except EmptyPeersListError:
                continue

    def shutdown(self):
        self._run = False
        self.socket_server.shutdown()



from threading import Thread
from typing import List
from abc import ABC, abstractmethod
from random import randrange

from .state import State
from .base_communication import BaseCommunication
from .heartbeat import Heartbeat
from .base_config import BaseConfig
from .job import Job


class BaseNode(ABC, Thread):
    def __init__(
            self,
            initial_config: BaseConfig,
            initial_state: State,
            communication: BaseCommunication,
    ):
        self.node_id = randrange(1, 9999999999)
        self.config = initial_config
        self.state = initial_state
        self.heartbeat = Heartbeat(interval=initial_config.heartbeat_interval)
        self.communication = communication

        super().__init__(name=f'Node {self.node_id}', daemon=True)

    @property
    def assigned_jobs_count(self) -> int:
        return len(list(filter(lambda job: job.assigned_to == self.node_id, self.state.assigned_jobs)))

    @abstractmethod
    def add_jobs(self):
        raise NotImplementedError

    @abstractmethod
    def assign_to_jobs(self, pending_jobs: List[Job]) -> List[Job]:
        raise NotImplementedError

    @abstractmethod
    def do_jobs(self, my_assigned_jobs: List[Job]):
        raise NotImplementedError

    @abstractmethod
    def completed_jobs(self, done_jobs: List[Job]):
        raise NotImplementedError

    def __merge_peers_state(self, peers_state: List[State]):
        for peer_state in peers_state:
            self.state.merge(peer_state)

    def __filter_my_assigned_jobs(self) -> List[Job]:
        return list(filter(lambda job: job.assigned_to == self.node_id, self.state.assigned_jobs))

    def __remove_done_jobs(self):
        for job in self.state.done_jobs:
            self.state.remove_job(job.id)

    def run(self):
        for beat_number in self.heartbeat:
            peers_state = self.communication.pull()
            self.__merge_peers_state(peers_state)

            self.completed_jobs(self.state.done_jobs)
            self.__remove_done_jobs()
            self.do_jobs(self.__filter_my_assigned_jobs())
            assigned_jobs = self.assign_to_jobs(self.state.pending_jobs)
            for job in assigned_jobs:
                job.assign(self.node_id)

            new_jobs = self.add_jobs()
            for new_job in new_jobs:
                self.state.add_job(new_job)

            self.communication.broadcast(self.state)



from .base_communication import BaseCommunication
from .state import State
from .base_config import BaseConfig
from .base_node import BaseNode
from .job import Job


__all__ = ["BaseCommunication", "State", 'BaseNode', "BaseConfig", "Job"]



from datetime import datetime
from json import dumps

from .state import State


class Message:
    def __init__(self, sender: str, created_at: datetime, state: State):
        self.sender = sender
        self.created_at = created_at
        self.state = state

    def to_json(self) -> str:
        return dumps({
            'sender': self.sender,
            'created_at': self.created_at.isoformat(),
            'state': self.state.to_dict()
        })

    @classmethod
    def from_dict(cls):
        pass



from setuptools import setup, find_packages

with open('README.md') as file:
    long_description = file.read()


setup(
      name='pyants',
      version='0.1.0',
      description='Allow you to create you own custom decentralize job management system.',
      url='https://github.com/hvushg/ants',
      author='Yehoyada.s',
      author_email='hvuhsg6@gmail.com',
      license='GPLv3',
      packages=find_packages(),
      keywords=['decentralize', 'job', 'jobs', 'job management'],
      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Natural Language :: English',
      ],
      long_description=long_description,
      long_description_content_type='text/markdown',
)



from typing import List
from abc import ABC, abstractmethod

from .state import State


class BaseCommunication(ABC):

    @abstractmethod
    def pull(self) -> List[State]:
        """ Pull sent data from other nodes
        :return: list of State objects
        """
        raise NotImplementedError

    @abstractmethod
    def broadcast(self, state):
        """ Send my state to peers
        can be all of them, just connected ones or even geo close ones (if using bluetooth)
        :param state: state of my node
        """
        raise NotImplementedError



from .job import JobStatus, Job


class State:
    def __init__(self, jobs: dict = None):
        if jobs is None:
            jobs = {}
        self._jobs = jobs  # {job_id: job_object}

    def add_job(self, job: Job):
        existing_job = self._jobs.get(job.id, None)
        if existing_job and existing_job > job:
            return
        self._jobs[job.id] = job

    def remove_job(self, job_id: str):
        del self._jobs[job_id]

    def _filter_jobs_by_status(self, status: JobStatus):
        return list(filter(lambda job: job.status == status, self._jobs.values()))

    @property
    def pending_jobs(self) -> list:
        return self._filter_jobs_by_status(JobStatus.PENDING)

    @property
    def assigned_jobs(self) -> list:
        return self._filter_jobs_by_status(JobStatus.ASSIGNED)

    @property
    def done_jobs(self) -> list:
        return self._filter_jobs_by_status(JobStatus.DONE)

    def merge(self, other_state):
        for job in other_state._jobs.values():
            my_job = self._jobs.get(job.id, None)
            if my_job:
                job = max(job, my_job)
            self._jobs[job.id] = job

    def __str__(self):
        state_str = 'State(\n'
        for job in self._jobs.values():
            state_str += f'{job}\n'
        state_str += ')'
        return state_str
